@@ -4,9 +4,13 @@ import copy
 import json
 import uuid
 import logging
+from packaging import version
 
 logger = logging.getLogger(__name__)
 
+# Check if 'aioredis' v2 is installed. TODO: remove support for 'aioredis' v1 when it is deprecated.
+aioredis_v2 = version.parse(aioredis.__version__) >= version.parse("2.0.0")
+
 
 class PlanQueueOperations:
     """
@@ -69,6 +73,11 @@ class PlanQueueOperations:
         self._name_plan_history = "plan_history"
         self._name_plan_queue_mode = "plan_queue_mode"
 
+        # Redis is also used for storage of some additional information not related to the queue.
+        #   The class contains only the functions for saving and retrieving the data, which is
+        #   not used by other functions of the class.
+        self._name_user_group_permissions = "user_group_permissions"
+
         # The list of allowed item parameters used for parameter filtering. Filtering operation
         #   involves removing all parameters that are not in the list.
         self._allowed_item_parameters = (
@@ -222,8 +231,13 @@ class PlanQueueOperations:
             self._lock = asyncio.Lock()
             async with self._lock:
                 try:
-                    self._r_pool = await aioredis.create_redis_pool(f"redis://{self._redis_host}", encoding="utf8")
-                except OSError as ex:
+                    host = f"redis://{self._redis_host}"
+                    if aioredis_v2:
+                        self._r_pool = aioredis.from_url(host, encoding="utf-8", decode_responses=True)
+                        await self._r_pool.ping()
+                    else:
+                        self._r_pool = await aioredis.create_redis_pool(host, encoding="utf8")
+                except Exception as ex:
                     error_msg = (
                         f"Failed to create the Redis pool: "
                         f"Redis server may not be available at '{self._redis_host}'. "
@@ -283,7 +297,8 @@ class PlanQueueOperations:
         async with self._lock:
             await self._delete_pool_entries()
 
-    def _verify_item_type(self, item):
+    @staticmethod
+    def _verify_item_type(item):
         """
         Check that the item (plan) is a dictionary.
         """
@@ -654,8 +669,8 @@ class PlanQueueOperations:
 
         Parameters
         ----------
-        plan: dict
-            Dictionary of plan parameters. Must be identical to the plan that is
+        item: dict
+            Dictionary of item parameters. Must be identical to the item that is
             expected to be deleted.
         single: boolean
             True - RuntimeError exception is raised if no or more than one matching
@@ -869,10 +884,16 @@ class PlanQueueOperations:
                     qsize = await self._r_pool.lpush(self._name_plan_queue, json.dumps(item))
             else:
                 item_to_displace = self._uid_dict_get_item(uid)
-                before = uid == before_uid
-                qsize = await self._r_pool.linsert(
-                    self._name_plan_queue, json.dumps(item_to_displace), json.dumps(item), before=before
-                )
+                if aioredis_v2:
+                    where = "BEFORE" if (uid == before_uid) else "AFTER"
+                    qsize = await self._r_pool.linsert(
+                        self._name_plan_queue, where, json.dumps(item_to_displace), json.dumps(item)
+                    )
+                else:
+                    before = uid == before_uid
+                    qsize = await self._r_pool.linsert(
+                        self._name_plan_queue, json.dumps(item_to_displace), json.dumps(item), before=before
+                    )
         elif pos == "back" or (isinstance(pos, int) and pos >= qsize0):
             qsize = await self._r_pool.rpush(self._name_plan_queue, json.dumps(item))
         elif pos == "front" or (isinstance(pos, int) and (pos == 0 or pos <= -qsize0)):
@@ -881,9 +902,14 @@ class PlanQueueOperations:
             # Put the position in the range
             item_to_displace = await self._get_item(pos=pos)
             if item_to_displace:
-                qsize = await self._r_pool.linsert(
-                    self._name_plan_queue, json.dumps(item_to_displace), json.dumps(item), before=True
-                )
+                if aioredis_v2:
+                    qsize = await self._r_pool.linsert(
+                        self._name_plan_queue, "BEFORE", json.dumps(item_to_displace), json.dumps(item)
+                    )
+                else:
+                    qsize = await self._r_pool.linsert(
+                        self._name_plan_queue, json.dumps(item_to_displace), json.dumps(item), before=True
+                    )
             else:
                 raise RuntimeError(f"Could not find an existing plan at {pos}. Queue size: {qsize0}")
         else:
@@ -1078,7 +1104,12 @@ class PlanQueueOperations:
 
         # Insert the new item after the old one and remove the old one. At this point it is guaranteed
         #   that they are not equal.
-        await self._r_pool.linsert(self._name_plan_queue, json.dumps(item_to_replace), json.dumps(item))
+        if aioredis_v2:
+            await self._r_pool.linsert(
+                self._name_plan_queue, "AFTER", json.dumps(item_to_replace), json.dumps(item)
+            )
+        else:
+            await self._r_pool.linsert(self._name_plan_queue, json.dumps(item_to_replace), json.dumps(item))
         await self._remove_item(item_to_replace)
 
         # Update self._uid_dict
@@ -1466,77 +1497,136 @@ class PlanQueueOperations:
                 del item["properties"]
         return item
 
-    async def _process_next_item(self):
+    async def _process_next_item(self, *, item=None):
         """
         See ``self.process_next_item()`` method.
         """
         loop_mode = self._plan_queue_mode["loop"]
 
         # Read the item from the front of the queue
-        item = await self._get_item(pos="front")
+
+        immediate_execution = bool(item)
+        if immediate_execution:
+            # Generate UID if it does not exist or creates a deep copy
+            item = copy.deepcopy(item) if "item_uid" in item else self.set_new_item_uuid(item)
+        else:
+            item = await self._get_item(pos="front")
+
         item_to_return = item
         if item:
             item_type = item["item_type"]
             if item_type == "plan":
-                item_to_return = await self._set_next_item_as_running()
-            else:
+                kwargs = {"item": item} if immediate_execution else {}
+                item_to_return = await self._set_next_item_as_running(**kwargs)
+
+            elif not immediate_execution:
                 # Items other than plans should be pushed to the back of the queue.
                 await self._pop_item_from_queue(pos="front")
                 if loop_mode:
                     item_to_add = self.set_new_item_uuid(item)
                     await self._add_item_to_queue(item_to_add)
+
         return item_to_return
 
-    async def process_next_item(self):
+    async def process_next_item(self, *, item=None):
         """
         Process the next item in the queue. If the item is a plan, it is set as currently
         running (``self.set_next_item_as_running``), otherwise it is popped from the queue.
         If the queue is LOOP mode, then it the item other than plan is pushed to the back
         of the queue. (Plans are pushed to the back of the queue upon successful completion.)
+
+        If ``item`` is a plan, then the plan is set for immediate execution. If ``item`` is
+        an instruction, then the function does nothing.
+
+        If an item submitted for 'immediate' execution has no UID, a new UID is created.
+        The returned plan is in the exact form, which is set for execution.
+
+        For more details on processing of queue plans, see the description for
+        ``self.set_next_item_as_running`` method.
         """
         async with self._lock:
-            return await self._process_next_item()
+            return await self._process_next_item(item=item)
 
-    async def _set_next_item_as_running(self):
+    async def _set_next_item_as_running(self, *, item=None):
         """
         See ``self.set_next_item_as_running()`` method.
         """
+        immediate_execution = bool(item)
+        if immediate_execution:
+            # Generate UID if it does not exist or creates a deep copy
+            item = copy.deepcopy(item) if "item_uid" in item else self.set_new_item_uuid(item)
+            item.setdefault("properties", {})["immediate_execution"] = True
+
         # UID remains in the `self._uid_dict` after this operation.
-        if not await self._is_item_running():
-            plan_json = await self._r_pool.lpop(self._name_plan_queue)
-            if plan_json:
-                plan = json.loads(plan_json)
-                if plan["item_type"] != "plan":
-                    raise RuntimeError(
-                        "Function 'PlanQueueOperations.set_next_item_as_running' was called for "
-                        f"an item other than plan: {plan}"
-                    )
-                self._plan_queue_uid = self.new_item_uid()
-                await self._set_running_item_info(plan)
+        try:
+            if await self._is_item_running():
+                raise Exception()
+
+            if immediate_execution:
+                plan = item
             else:
-                plan = {}
-        else:
+                plan = await self._get_item(pos="front")
+                if not plan:
+                    raise Exception()
+
+            if "item_type" not in plan:
+                raise Exception()
+
+            if plan["item_type"] != "plan":
+                raise RuntimeError(
+                    "Function 'PlanQueueOperations.set_next_item_as_running' was called for "
+                    f"an item other than plan: {plan}"
+                )
+
+            if not immediate_execution:
+                # Pop plan from the front of the queue (it is the same plan as currently loaded)
+                await self._r_pool.lpop(self._name_plan_queue)
+                self._plan_queue_uid = self.new_item_uid()
+
+            await self._set_running_item_info(plan)
+
+        except RuntimeError:
+            raise
+        except Exception:
             plan = {}
+
         return plan
 
-    async def set_next_item_as_running(self):
+    async def set_next_item_as_running(self, *, item=None):
         """
         Sets the next item from the queue as 'running'. The item MUST be a plan
         (e.g. not an instruction), otherwise an exception will be raised. The item is removed
         from the queue. UID remains in ``self._uid_dict``, i.e. item with the same UID
-        may not be added to the queue while it is being executed.
+        may not be added to the queue while it is being executed. If ``item`` parameter
+        represents a plan, then the plan is set for immediate execution.
 
         This function can only be applied to the plans. Use ``process_next_item`` that
         works correctly for any item (it calls this function if an item is a plan).
 
+        If a plan submitted for 'immediate' execution has no UID, a new UID is created.
+        The returned plan is in the exact form, which is set for execution.
+
+        Parameters
+        ----------
+        item: dict or None
+            The dictionary that represents a plan submitted for immediate execution.
+            If ``item`` is a plan, then the queue remains intact and the plan is
+            set for immediate execution. If ``item=None``, then the top queue item
+            is removed from the queue and set for execution.
+
         Returns
         -------
         dict
             The item that was set as currently running. If another item is currently
             set as 'running' or the queue is empty, then ``{}`` is returned.
+
+        Raises
+        ------
+        RuntimeError
+            The function is called for an item other than plan.
         """
         async with self._lock:
-            return await self._set_next_item_as_running()
+            return await self._set_next_item_as_running(item=item)
 
     async def _set_processed_item_as_completed(self, exit_status, run_uids):
         """
@@ -1560,7 +1650,7 @@ class PlanQueueOperations:
             item_cleaned["result"]["exit_status"] = exit_status
             item_cleaned["result"]["run_uids"] = run_uids
             await self._clear_running_item_info()
-            if not loop_mode:
+            if not loop_mode and not immediate_execution:
                 self._uid_dict_remove(item["item_uid"])
             await self._add_to_history(item_cleaned)
         else:
@@ -1638,3 +1728,35 @@ class PlanQueueOperations:
         """
         async with self._lock:
             return await self._set_processed_item_as_stopped(exit_status=exit_status, run_uids=run_uids)
+
+    # =============================================================================================
+    #         Methods for saving and retrieving user group permissions.
+
+    async def user_group_permissions_clear(self):
+        """
+        Clear user group permissions saved in Redis.
+        """
+        await self._r_pool.delete(self._name_user_group_permissions)
+
+    async def user_group_permissions_save(self, user_group_permissions):
+        """
+        Save user group permissions to Redis.
+
+        Parameters
+        ----------
+        user_group_permissions: dict
+            A dictionary containing user group permissions.
+        """
+        await self._r_pool.set(self._name_user_group_permissions, json.dumps(user_group_permissions))
+
+    async def user_group_permissions_retrieve(self):
+        """
+        Retreive saved user group permissions.
+
+        Returns
+        -------
+        dict or None
+            Returns dictionary with saved user group permissions or ``None`` if no permissions are saved.
+        """
+        ugp_json = await self._r_pool.get(self._name_user_group_permissions)
+        return json.loads(ugp_json) if ugp_json else None
