@@ -1,6 +1,8 @@
 import time as ttime
 import subprocess
 import pytest
+import os
+import yaml
 
 from bluesky_queueserver.manager.profile_ops import gen_list_of_plans_and_devices
 from bluesky_queueserver.manager.comms import generate_new_zmq_key_pair
@@ -17,9 +19,9 @@ from .common import (
     get_reduced_state_info,
     get_queue_state,
     get_queue,
-    copy_default_profile_collection,
     append_code_to_last_startup_file,
     set_qserver_zmq_public_key,
+    zmq_single_request,
 )
 
 from .common import re_manager, re_manager_pc_copy, re_manager_cmd  # noqa: F401
@@ -33,6 +35,8 @@ REQ_FAILED = QServerExitCodes.REQUEST_FAILED.value
 COM_ERROR = QServerExitCodes.COMMUNICATION_ERROR.value
 EXCEPTION_OCCURRED = QServerExitCodes.EXCEPTION_OCCURRED.value
 
+timeout_env_open = 10
+
 
 def test_qserver_cli_and_manager(re_manager):  # noqa: F811
     """
@@ -49,6 +53,10 @@ def test_qserver_cli_and_manager(re_manager):  # noqa: F811
     assert subprocess.call(["qserver", "allowed", "plans"], stdout=subprocess.DEVNULL) == SUCCESS
     assert subprocess.call(["qserver", "allowed", "devices"], stdout=subprocess.DEVNULL) == SUCCESS
 
+    # Request the list of existing plans and devices (we don't check what is returned)
+    assert subprocess.call(["qserver", "existing", "plans"], stdout=subprocess.DEVNULL) == SUCCESS
+    assert subprocess.call(["qserver", "existing", "devices"], stdout=subprocess.DEVNULL) == SUCCESS
+
     # Add a number of plans
     plan_1 = "{'name':'count', 'args':[['det1', 'det2']]}"
     plan_2 = "{'name':'scan', 'args':[['det1', 'det2'], 'motor', -1, 1, 10]}"
@@ -68,15 +76,11 @@ def test_qserver_cli_and_manager(re_manager):  # noqa: F811
     assert n_plans == 2, "Incorrect number of plans in the queue"
 
     assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
-    assert wait_for_condition(
-        time=3, condition=condition_environment_created
-    ), "Timeout while waiting for environment to be created"
+    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)
 
     assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
 
-    assert wait_for_condition(
-        time=60, condition=condition_queue_processing_finished
-    ), "Timeout while waiting for process to finish"
+    assert wait_for_condition(time=60, condition=condition_queue_processing_finished)
 
     # Smoke test for 'history_get' and 'history_clear'
     assert subprocess.call(["qserver", "history", "get"]) == SUCCESS
@@ -162,9 +166,7 @@ def test_qserver_environment_close(re_manager):  # noqa: F811
     assert is_plan_running is False
 
     assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
-    assert wait_for_condition(
-        time=3, condition=condition_environment_created
-    ), "Timeout while waiting for environment to be opened"
+    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)
 
     assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
     ttime.sleep(2)
@@ -210,9 +212,7 @@ def test_qserver_environment_destroy(re_manager):  # noqa: F811
     assert is_plan_running is False
 
     assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
-    assert wait_for_condition(
-        time=3, condition=condition_environment_created
-    ), "Timeout while waiting for environment to be opened"
+    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)
 
     assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
     ttime.sleep(2)
@@ -230,9 +230,7 @@ def test_qserver_environment_destroy(re_manager):  # noqa: F811
     assert is_plan_running is False
 
     assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
-    assert wait_for_condition(
-        time=3, condition=condition_environment_created
-    ), "Timeout while waiting for environment to be opened"
+    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)
 
     assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
     ttime.sleep(2)
@@ -287,10 +285,8 @@ def test_qserver_re_pause_continue(re_manager, option_pause, option_continue):
     assert n_plans == 2, "Incorrect number of plans in the queue"
     assert is_plan_running is False
 
-    assert subprocess.call(["qserver", "environment", "open"]) == 0
-    assert wait_for_condition(
-        time=3, condition=condition_environment_created
-    ), "Timeout while waiting for environment to be opened"
+    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
+    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)
 
     assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
     ttime.sleep(2)
@@ -383,9 +379,7 @@ def test_qserver_manager_kill(re_manager, time_kill):  # noqa: F811
     assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS
 
     assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
-    assert wait_for_condition(
-        time=3, condition=condition_environment_created
-    ), "Timeout while waiting for environment to be opened"
+    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)
 
     if time_kill == "before":
         # The command that kills manager always times out
@@ -466,7 +460,7 @@ def test_qserver_env_open_various_cases(re_manager_pc_copy, additional_code, suc
     # Wait until RE Manager is started
     assert wait_for_condition(time=10, condition=condition_manager_idle)
 
-    # Attempt to create the environment
+    # Attempt to create the environment. Long timeout.
     assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
     assert wait_for_condition(time=30, condition=condition_manager_idle)
 
@@ -477,7 +471,7 @@ def test_qserver_env_open_various_cases(re_manager_pc_copy, additional_code, suc
         # Remove the offending patch and try to start the environment again. It should work
         patch_first_startup_file_undo(pc_path)
         assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
-        assert wait_for_condition(time=3, condition=condition_environment_created)
+        assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)
 
     # Run a plan to make sure RE Manager is functional after the startup.
     plan = "{'name':'count', 'args':[['det1', 'det2']], 'kwargs':{'num': 10, 'delay': 1}}"
@@ -511,7 +505,7 @@ def test_qserver_manager_stop_1(re_manager, option):  # noqa: F811
 
     # Attempt to create the environment
     assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
-    assert wait_for_condition(time=30, condition=condition_manager_idle)
+    assert wait_for_condition(time=timeout_env_open, condition=condition_manager_idle)
 
     cmd = ["qserver", "manager", "stop"]
     if option:
@@ -535,7 +529,7 @@ def test_qserver_manager_stop_2(re_manager, option):  # noqa: F811
 
     # Attempt to create the environment
     assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
-    assert wait_for_condition(time=30, condition=condition_manager_idle)
+    assert wait_for_condition(time=timeout_env_open, condition=condition_manager_idle)
 
     plan = "{'name':'count', 'args':[['det1', 'det2']], 'kwargs':{'num': 10, 'delay': 1}}"
     assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS
@@ -868,9 +862,7 @@ def test_qserver_item_execute_1(re_manager, item_type, env_exists):  # noqa: F81
 
     if env_exists:
         assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
-        assert wait_for_condition(
-            time=3, condition=condition_environment_created
-        ), "Timeout while waiting for environment to be created"
+        assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)
 
     expected_result = SUCCESS if env_exists else REQ_FAILED
     expected_n_history = 1 if env_exists and (item_type == "plan") else 0
@@ -1041,7 +1033,7 @@ def test_qserver_queue_stop(re_manager, deactivate):  # noqa: F811
 
     # Attempt to create the environment
     assert subprocess.call(["qserver", "environment", "open"]) == 0
-    assert wait_for_condition(time=10, condition=condition_manager_idle)
+    assert wait_for_condition(time=timeout_env_open, condition=condition_manager_idle)
 
     plan = "{'name':'count', 'args':[['det1', 'det2']], 'kwargs':{'num': 10, 'delay': 1}}"
     assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == 0
@@ -1108,6 +1100,137 @@ def test_qserver_re_runs(re_manager, option, exit_code):  # noqa: F811
     assert subprocess.call(["qserver", "re", "runs", *params]) == exit_code
 
 
+_script_to_upload_1 = """
+def count_modified(detectors, num=1, delay=None):
+    yield from count(detectors=detectors, num=num, delay=delay)
+"""
+
+
+# fmt: off
+@pytest.mark.parametrize("run_in_background", [False, True])
+@pytest.mark.parametrize("update_re", [False, True])
+# fmt: on
+def test_script_upload_1(re_manager, tmp_path, run_in_background, update_re):  # noqa: F811
+    """
+    Tests for 'qserver script upload'. The uploaded script does not change RE, so the tests
+    simply checks if 'update-re' parameter is accepted.
+    """
+    # Create a file with the script
+    script_fln = "script_to_upload.py"
+    script_path = os.path.join(tmp_path, script_fln)
+    with open(script_path, "w") as f:
+        f.writelines(_script_to_upload_1)
+
+    params = ["qserver", "script", "upload", script_path]
+    if run_in_background:
+        params.append("background")
+    if update_re:
+        params.append("update-re")
+
+    # Call is expected to fail (environment is not open)
+    assert subprocess.call(params) == REQ_FAILED
+
+    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
+    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)
+
+    # Check that 'count_modified' plan does not exist before the script is uploaded
+    resp1, _ = zmq_single_request("plans_existing")
+    assert resp1["success"] is True
+    assert "count_modified" not in resp1["plans_existing"]
+
+    if run_in_background:
+        # Start some foreground process (run a function)
+        item = r"{'name': 'function_sleep', 'kwargs': {'time': 2}}"
+        params_fg = ["qserver", "function", "execute", item]
+        assert subprocess.call(params_fg) == SUCCESS
+
+    assert subprocess.call(params) == SUCCESS
+    if run_in_background:
+        ttime.sleep(2)  # Wait for the task to be completed
+    assert wait_for_condition(time=5, condition=condition_manager_idle)
+
+    # Check that 'count_modified' function was loaded into the environment
+    resp2, _ = zmq_single_request("plans_existing")
+    assert resp2["success"] is True
+    assert "count_modified" in resp2["plans_existing"]
+
+    assert subprocess.call(["qserver", "environment", "close"]) == SUCCESS
+    assert wait_for_condition(time=5, condition=condition_environment_closed)
+
+
+def test_script_upload_2_fail(re_manager, tmp_path):  # noqa: F811
+    """
+    Tests for 'qserver script upload': failing cases.
+    """
+    script_path = os.path.join(tmp_path, "script.py")
+    with open(script_path, "w") as f:
+        f.writelines(_script_to_upload_1)
+
+    # File does not exist (IO error)
+    assert subprocess.call(["qserver", "script", "upload", os.path.join(tmp_path, "script2.py")]) == PARAM_ERROR
+
+    # Invalid parameter
+    assert subprocess.call(["qserver", "script", "upload", script_path, "invalid_param"]) == PARAM_ERROR
+
+
+# fmt: off
+@pytest.mark.parametrize("run_in_background", [False, True])
+# fmt: on
+def test_function_execute_1(re_manager, run_in_background):  # noqa: F811
+    """
+    Tests for 'qserver function execute'.
+    """
+    item = r"{'name': 'function_sleep', 'kwargs': {'time': 2}}"
+    params = ["qserver", "function", "execute", item]
+    params_fg = params.copy()  # Parameters for foreground task
+    if run_in_background:
+        params.append("background")
+
+    # Call is expected to fail (environment is not open)
+    assert subprocess.call(params) == REQ_FAILED
+
+    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
+    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)
+
+    if run_in_background:
+        # Start a foreground task (the same function) to test if a function can be run in the background
+        assert subprocess.call(params_fg) == SUCCESS
+        # Attempt to start another foreground process
+        assert subprocess.call(params_fg) == REQ_FAILED
+
+    assert subprocess.call(params) == SUCCESS
+    if run_in_background:
+        ttime.sleep(3)  # Wait for the task to be completed
+    assert wait_for_condition(time=5, condition=condition_manager_idle)
+
+    assert subprocess.call(["qserver", "environment", "close"]) == SUCCESS
+    assert wait_for_condition(time=5, condition=condition_environment_closed)
+
+
+def test_function_execute_2_fail(re_manager):  # noqa: F811
+    """
+    Tests for 'qserver function execute': failing cases.
+    """
+    item = r"{'name': 'function_sleep', 'kwargs': {'time': 2}}"
+
+    # Invalid parameter
+    assert subprocess.call(["qserver", "function", "execute", item, "invalid_param"]) == PARAM_ERROR
+
+
+def test_task_result_1(re_manager):  # noqa: F811
+    """
+    Tests for 'qserver task result_get'.
+    """
+    # The request should be successful for any 'task_uid'.
+    task_uid = "01e80342-5e36-44de-bc86-9bd8d57c9885"
+    assert subprocess.call(["qserver", "task", "result", task_uid]) == SUCCESS
+
+    # Some cases of invalid parameters
+    assert subprocess.call(["qserver", "task", "result"]) == PARAM_ERROR
+    assert subprocess.call(["qserver", "task", "something"]) == PARAM_ERROR
+    assert subprocess.call(["qserver", "task", "result", task_uid, "extra_param"]) == PARAM_ERROR
+
+
 _sample_trivial_plan1 = """
 def trivial_plan_for_unit_test():
     '''
@@ -1117,11 +1240,15 @@ def trivial_plan_for_unit_test():
 """
 
 
-def test_qserver_reload_permissions(re_manager_pc_copy, tmp_path):  # noqa F811
+# fmt: off
+@pytest.mark.parametrize("reload_plans_devices", [True, False])
+# fmt: on
+def test_qserver_permissions_reload_1(re_manager_pc_copy, tmp_path, reload_plans_devices):  # noqa F811
     """
-    Tests for ``/permissions/reload`` API.
+    Tests for ``qserver permissions reload [lists]`` API.
     """
-    pc_path = copy_default_profile_collection(tmp_path)
+    # pc_path = copy_default_profile_collection(tmp_path)
+    _, pc_path = re_manager_pc_copy
     append_code_to_last_startup_file(pc_path, additional_code=_sample_trivial_plan1)
 
     # Generate the new list of allowed plans and devices and reload them
@@ -1134,10 +1261,68 @@ def test_qserver_reload_permissions(re_manager_pc_copy, tmp_path):  # noqa F811
     assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == REQ_FAILED
 
     # Reload profile collection
-    assert subprocess.call(["qserver", "permissions", "reload"]) == SUCCESS
+    params = ["permissions", "reload"]
+    if reload_plans_devices:
+        params.append("lists")
+    assert subprocess.call(["qserver", *params]) == SUCCESS
 
     # Attempt to add the plan to the queue. It should be successful now.
-    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS
+    res_expected = SUCCESS if reload_plans_devices else REQ_FAILED
+    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == res_expected
+
+
+_permissions_dict_not_allow_count = {
+    "user_groups": {
+        "root": {"allowed_plans": [None], "allowed_devices": [None]},
+        "admin": {"allowed_plans": [None], "forbidden_plans": ["^count$"], "allowed_devices": [None]},
+    }
+}
+
+
+def test_qserver_permissions_set_get_1(re_manager, tmp_path):  # noqa F811
+    """
+    Tests for ``qserver permissions set`` and ``qserver permissions get``: basic tests.
+    """
+
+    user_group_permissions = _permissions_dict_not_allow_count
+
+    # Create yaml file with user group permissions
+    fl_path = os.path.join(tmp_path, "ugp.yaml")
+    with open(fl_path, "w") as f:
+        f.writelines(yaml.dump(user_group_permissions))
+
+    # Set user group permissions
+    assert subprocess.call(["qserver", "permissions", "set", fl_path]) == SUCCESS
+
+    # Check that permissions were really changed
+    resp1, _ = zmq_single_request("permissions_get")
+    assert resp1["success"] is True, f"resp={resp1}"
+    assert resp1["user_group_permissions"] == user_group_permissions
+
+    # Check if 'qserver permissions get' works
+    assert subprocess.call(["qserver", "permissions", "get"]) == SUCCESS
+
+
+def test_qserver_permissions_set_get_2_fail(re_manager, tmp_path):  # noqa F811
+    """
+    Tests for ``qserver permissions set`` and ``qserver permissions get``: failing cases.
+    """
+    # Create yaml file with user group permissions
+    fl_path = os.path.join(tmp_path, "ugp.yaml")
+    with open(fl_path, "w") as f:
+        f.writelines("This is not a valid permissions dictionary")  # Invalid file
+
+    # Non-existing file
+    assert subprocess.call(["qserver", "permissions", "set", os.path.join(tmp_path, "some_file")]) == PARAM_ERROR
+    # No file name
+    assert subprocess.call(["qserver", "permissions", "set"]) == PARAM_ERROR
+    # Extra parameters
+    assert subprocess.call(["qserver", "permissions", "set", fl_path, "extra_parameter"]) == PARAM_ERROR
+    # Request is correct, but the file does not contain valid dictionary
+    assert subprocess.call(["qserver", "permissions", "set", fl_path]) == REQ_FAILED
+
+    # Extra parameters
+    assert subprocess.call(["qserver", "permissions", "get", "extra_parameter"]) == PARAM_ERROR
 
 
 # fmt: off
@@ -1176,10 +1361,12 @@ def test_qserver_secure_1(monkeypatch, re_manager_cmd, test_mode):  # noqa: F811
     assert subprocess.call(["qserver", "allowed", "plans"], stdout=subprocess.DEVNULL) == SUCCESS
     assert subprocess.call(["qserver", "allowed", "devices"], stdout=subprocess.DEVNULL) == SUCCESS
 
+    # Request the list of existing plans and devices (we don't check what is returned)
+    assert subprocess.call(["qserver", "existing", "plans"], stdout=subprocess.DEVNULL) == SUCCESS
+    assert subprocess.call(["qserver", "existing", "devices"], stdout=subprocess.DEVNULL) == SUCCESS
+
     assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
-    assert wait_for_condition(
-        time=3, condition=condition_environment_created
-    ), "Timeout while waiting for environment to be created"
+    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)
 
     state = get_queue_state()
     assert state["items_in_queue"] == 2
@@ -1216,4 +1403,5 @@ def test_qserver_zmq_keys():
 
     # Generated public key based on private key - success
     _, private_key = generate_new_zmq_key_pair()
+    print(f"Private key used for the test: '{private_key}'")
     assert subprocess.call(["qserver-zmq-keys", "--zmq-private-key", private_key]) == SUCCESS
